from typing import NamedTuple


class DOI(NamedTuple):
    prefix: str
    suffix: str

    def __str__(self):
        return self.to_str()

    def to_str(self):
        return "doi:" + self.to_minimal()

    def to_minimal(self):
        return f"{self.prefix}/{self.suffix}"

    def to_url(self):
        return "https://doi.org/" + self.to_minimal()

    def to_handle(self):
        return "info:doi/" + self.to_minimal()

    @classmethod
    def parse(cls, s: str):
        s = s.strip()
        skip = 0
        if s.startswith("doi:"):
            skip = len("doi:")
        elif s.startswith("http"):
            skip = len("http://doi.org/")
            if s[4] == "s":
                skip += 1
        elif s.startswith("info:doi/"):
            skip = len("info:doi/")
        pref_suf = s[skip:].split("/")
        if len(pref_suf) != 2:
            raise ValueError(f"Could not parse DOI from {pref_suf}")
        return cls(*pref_suf)



import datetime as dt
import typing as tp
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from urllib.request import Request

from ..doi import DOI


class UpdateType(Enum):
    RETRACTION = "RETRACTION"
    CONCERN = "CONCERN"
    UPDATE = "UPDATE"


@dataclass
class Update:
    original_doi: DOI
    type: UpdateType
    timestamp: tp.Optional[dt.datetime] = None
    update_doi: tp.Optional[DOI] = None


class Backend(ABC):
    @abstractmethod
    def make_request(self, doi: DOI) -> Request:
        pass

    @abstractmethod
    def parse_content(self, content: bytes) -> tp.Optional[Update]:
        pass



import datetime as dt
import json
import typing as tp
from dataclasses import dataclass
from urllib.request import Request

from ..doi import DOI
from .base import Backend, Update, UpdateType

Jso = tp.Optional[tp.Union[tp.Dict[str, "Jso"], tp.List["Jso"], int, float, str]]

OPENRETRACTIONS_URL_FMT = "http://openretractions.com/api/doi/{}/data.json"


@dataclass
class ORUpdate:
    timestamp: dt.datetime
    doi: DOI
    type: str

    @classmethod
    def from_jso(cls, jso: tp.Dict[str, Jso]):
        return cls(
            dt.datetime.fromtimestamp(jso["timestamp"] / 1000),
            DOI.parse(jso["identifier"]["doi"]),
            str(jso["type"]),
        )


@dataclass
class OREntry:
    """
    Note that this documentation is wrong, just copied from the site.

    {
        "retracted": false,  // whether or not the paper has been retracted
        // the UNIXtime when the original paper was published
        "timestamp": 1361836800000,
        "update": {
            "timestamp": 1491464002919, // the UNIXtime when the update was recorded
            "doi": "10.1002/job.1858", // the DOI of the update
            "type": "correction" // the publisher's description of the update
        },
        "doi": "10.1002/job.1787", // the DOI of the original paper
        "journal": "Journal of Organizational Behavior",
        "publisher": "Wiley-Blackwell",
        "title": "Erratum: Cognitive and affective identification:
            Exploring the links between different forms of social "
            "identification and personality with work attitudes and behavior"
    }
    """

    retracted: bool
    timestamp: dt.datetime
    updates: tp.List[ORUpdate]
    doi: DOI
    journal: str
    publisher: tp.Optional[str]
    title: str

    @classmethod
    def from_jso(cls, jso: tp.Dict[str, Jso]):
        return cls(
            bool(jso["retracted"]),
            dt.datetime.fromtimestamp(jso["timestamp"] / 1000),
            [ORUpdate.from_jso(u) for u in jso["updates"]],
            DOI.parse(jso["identifier"]["doi"]),
            jso.get("journal"),
            jso["publisher"],
            jso["title"],
        )


def make_url(doi: tp.Union[str, DOI]):
    if isinstance(doi, str):
        doi = DOI.parse(doi)
    return OPENRETRACTIONS_URL_FMT.format(doi.to_minimal())


def parse_content(s: str):
    jso = json.loads(s)
    return OREntry.from_jso(jso)


class OpenRetractions(Backend):
    def make_request(self, doi: DOI) -> Request:
        if isinstance(doi, str):
            doi = DOI.parse(doi)
        url = OPENRETRACTIONS_URL_FMT.format(doi.to_minimal())
        return Request(url)

    def parse_content(self, content: bytes) -> tp.Optional[Update]:
        jso = json.loads(content.decode("utf-8"))
        or_entry = OREntry.from_jso(jso)
        if or_entry.retracted:
            return Update(or_entry.doi, UpdateType.RETRACTION)
        elif not or_entry.updates:
            return None

        last_update = or_entry.updates[-1]
        return Update(
            or_entry.doi,
            UpdateType(last_update.type),
            last_update.timestamp,
            last_update.doi,
        )



import asyncio
import logging
import sys
import typing as tp
from argparse import ArgumentParser
from pathlib import Path
from urllib.request import Request

import httpx

from .backends import UpdateType, registry
from .doi import DOI

logger = logging.getLogger(__name__)


def path_or_pipe(s: str) -> tp.Optional[Path]:
    if s == "-":
        return None
    return Path(s).resolve()


def read_dois(p: tp.Optional[Path]) -> tp.Iterator[DOI]:
    if p is None:
        for line in sys.stdin:
            yield DOI.parse(line)
    else:
        with open(p) as f:
            for line in f:
                yield DOI.parse(line)


async def fetch(reqs: tp.List[Request]) -> tp.List[tp.Optional[bytes]]:
    contents = dict()
    async with httpx.AsyncClient() as client:

        async def get(idx, req: Request):
            got = await client.request(
                req.get_method(),
                req.full_url,
                headers=dict(req.headers),
                # todo: handle POST request data
            )
            return idx, got

        for coro in asyncio.as_completed(get(idx, url) for idx, url in enumerate(reqs)):
            idx, r = await coro
            if r.status_code == 404:
                contents[idx] = None
                continue
            r.raise_for_status()
            contents[idx] = r.content

    return [contents[k] for k in sorted(contents)]


def main(args=None):
    parser = ArgumentParser()
    parser.add_argument(
        "doi",
        nargs="*",
        help="DOI to check for retractions or updates. Handled after --infile lists.",
    )
    parser.add_argument(
        "-r",
        "--retractions-only",
        action="store_true",
        help="Only list retracted DOIs, not updated",
    )
    backends = sorted(registry)
    parser.add_argument("-b", "--backend", choices=backends, default="openretractions")
    parser.add_argument(
        "-i",
        "--infile",
        type=path_or_pipe,
        action="append",
        help="Get DOIs from file (newline-separated); accepts multiple",
    )
    parser.add_argument(
        "-o", "--outfile", type=path_or_pipe, help="Write to file instead of stdout"
    )
    parser.add_argument("-u", "--url", action="store_true", help="Show DOIs as URLs")

    parsed = parser.parse_args(args)
    backend = registry[parsed.backend]

    def fmt_doi(doi: DOI):
        if parsed.url:
            return doi.to_url()
        else:
            doi.to_str()

    requests = []
    if parsed.infile:
        for infile in parsed.infile:
            requests.extend(backend.make_request(d) for d in read_dois(infile))
    requests.extend(backend.make_request(d) for d in parsed.doi)

    response_contents = asyncio.run(fetch(requests))
    rows = []
    for content in response_contents:
        if content is None:
            continue
        update = backend.parse_content(content)
        if update is None:
            continue
        row = [update.type.value]
        row.append(fmt_doi(update.original_doi))

        if not update.type == UpdateType.RETRACTION:
            if parsed.retractions_only:
                continue

            row.append(fmt_doi(update.update_doi) if update.update_doi else "")
            row.append(update.timestamp.isoformat() if update.timestamp else "")

        rows.append("\t".join(row))

    if parsed.outfile:
        with open(parsed.outfile, "w") as f:
            for row in rows:
                f.write(row + "\n")
    else:
        for row in rows:
            print(row)



from pathlib import Path

from setuptools import find_packages, setup

with open(Path(__file__).resolve().parent / "README.md") as f:
    readme = f.read()

extras = {
    "cli": ["httpx"],
}

setup(
    name="retractions",
    url="https://github.com/clbarnes/retractions",
    author="Chris L. Barnes",
    description="Check DOIs for retractions",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(include=["retractions"]),
    install_requires=[],
    extras_require=extras,
    python_requires=">=3.8, <4.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    entry_points={"console_scripts": ["retractions=retractions.cli:main"]},
)
